(******************************************************************************

 ProtoVeriPhy

 Module: distance/MessageTheory3.thy

 Copyright (C) 2011-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)


theory MessageTheoryXor3 imports MessageTheoryXor2 begin


fun
  ffactors :: "fmsg \<Rightarrow> fmsg set"
 where
   "ffactors (XOR a b) = ffactors a \<union> ffactors b"
 | "ffactors (a) = {a}"

definition
 factors :: "msg \<Rightarrow> msg set"
 where
  "factors m \<equiv> {Abs_msg a | a . a \<in> ffactors (Rep_msg m)}"

inductive
  out_context :: "msg \<Rightarrow> msg \<Rightarrow> msg \<Rightarrow> bool"
 where
  Base[intro]:  "\<lbrakk> t = m; c \<noteq> m \<rbrakk>                     \<Longrightarrow> out_context t c m"           |
  Hash[intro]:  "\<lbrakk> out_context t c X; c \<noteq> Hash X \<rbrakk>    \<Longrightarrow> out_context t c (Hash X)"    |
  Crypt[intro]: "\<lbrakk> out_context t c X; c \<noteq> Crypt k X \<rbrakk> \<Longrightarrow> out_context t c (Crypt k X)" |
  PairL[intro]: "\<lbrakk> out_context t c X; c \<noteq> \<lbrace> X, Y\<rbrace> \<rbrakk>   \<Longrightarrow> out_context t c (\<lbrace>X, Y\<rbrace>)"    |
  PairR[intro]: "\<lbrakk> out_context t c Y; c \<noteq> \<lbrace> X, Y\<rbrace> \<rbrakk>   \<Longrightarrow> out_context t c (\<lbrace>X, Y\<rbrace>)"    |
  Xor[intro]:   "\<lbrakk> out_context t c m; m \<in> factors X; m \<noteq> X ; c \<noteq> X \<rbrakk>
                 \<Longrightarrow> out_context t c X"

(* The last case allows us to go from m to X = ... \<oplus> m \<oplus> ... if X \<noteq> c.
   Note that out_context NA (NB \<oplus> NA) X for X = NA \<oplus> NB \<oplus> NC holds with
   respect to this definition. So using an Xor as a context might yield unexpected
   results.
*)

lemma out_context_inverse:
  "out_context t c m
   \<Longrightarrow> m \<noteq> c
       \<and> ( m = t
          \<or> (\<exists> X. m = Hash X \<and> out_context t c X)
          \<or> (\<exists> k X. m = Crypt k X \<and> out_context t c X)
          \<or> (\<exists> X Y. m = \<lbrace>X, Y\<rbrace> \<and> (out_context t c X \<or> out_context t c Y))
          \<or> (\<exists> X \<in> factors m. m \<noteq> X \<and> (out_context t c X)))"
  apply (induct rule: out_context.induct)
  by auto

lemma out_context_nonce[simp]: "out_context (Nonce A NA) (Hash (Nonce A NA)) (Nonce A NA)"
  by auto

lemma "\<not> (out_context (Nonce A NA) (Hash (Nonce A NA)) (Hash (Nonce A NA)))"
  apply auto
  apply (drule out_context_inverse)
  apply auto
  done


lemma factors_Agent[simp]: "factors (Agent a) = {Agent a}"
  apply (subgoal_tac "AGENT a \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Agent_def factors_def Abs_msg_inverse)
done

lemma factors_Zero[simp]: "factors (Zero) = {Zero}"
  apply (subgoal_tac "ZERO \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Zero_def factors_def Abs_msg_inverse)
done


lemma factors_Real[simp]: "factors (Real a) = {Real a}"
  apply (subgoal_tac "REAL a \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Real_def factors_def Abs_msg_inverse)
done

lemma factors_Number[simp]: "factors (Number n) = {Number n}"
  apply (subgoal_tac "NUMBER n \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Number_def factors_def Abs_msg_inverse)
done

lemma factors_Nonce[simp]: "factors (Nonce A NA) = {Nonce A NA}"
  apply (subgoal_tac "NONCE A NA \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Nonce_def factors_def Abs_msg_inverse)
done


lemma factors_Key[simp]: "factors (Key k) = {Key k}"
  apply (subgoal_tac "KEY k \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Key_def factors_def Abs_msg_inverse)
done

lemma factors_Hash[simp]: "factors (Hash m) = {Hash m}"
  apply (subgoal_tac "HASH (Rep_msg m) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Hash_def factors_def Abs_msg_inverse)
  done

lemma factors_MPair[simp]: "factors \<lbrace>A,B\<rbrace> =  {\<lbrace>A,B\<rbrace>}"
  apply (subgoal_tac "MPAIR (Rep_msg A) (Rep_msg B) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: MPair_def factors_def Abs_msg_inverse)
  done

lemma factors_Crypt[simp]: "factors (Crypt K X) =  {Crypt K X}"
  apply (subgoal_tac "CRYPT K (Rep_msg X) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Crypt_def factors_def Abs_msg_inverse)
  done

lemma ffactors_fsubterms:
  "\<lbrakk>normed y; a \<in> ffactors y\<rbrakk> \<Longrightarrow> a \<in> fsubterms {y}"
  apply (induct rule: normed.induct)
  apply auto
  apply (erule contrapos_np) back
  apply (rule fsubterms.mono_elem)
  apply auto
  apply (erule contrapos_np) back
  apply (rule fsubterms.mono_elem)
  apply auto
  done

lemma factors_subset_subterms:
  "factors t \<subseteq> subterms {t}"
  apply (case_tac t)
  apply (auto simp add: factors_def subterms_def msg_def)
  apply (rule_tac x=a in exI)
  apply clarsimp
  apply (erule ffactors_fsubterms)
  by auto

lemma factors_imp_subterms: "a \<in> factors b \<Longrightarrow> a \<in> subterms {b}"
  apply (insert factors_subset_subterms[where t=b])
  by auto
  
lemma out_context_imp_subterms:
  "out_context t c m \<Longrightarrow> t \<in> subterms {m}"
  apply (erule out_context.induct)
  apply auto
  apply (drule factors_imp_subterms)
  apply (drule subterms.trans)
  apply auto
  done

lemma ffactors_xor_red:
  "x ~> y \<Longrightarrow> (\<forall> t. t \<in> ffactors y \<longrightarrow> ((\<exists> t'. ((t' \<approx> t) \<and> t' \<in> ffactors x)) \<or> t \<approx> ZERO))"
  apply (erule xor_red.induct)
  apply force
  apply force
  apply force
  apply force
  apply force
  apply auto

  apply (rule xor_eq.MPair_cong)
  apply (auto dest: xor_red_imp_xor_eq)

  apply (rule xor_eq.Hash_cong)
  apply (auto dest: xor_red_imp_xor_eq)

  apply (rule xor_eq.Crypt_cong)
  apply (auto dest: xor_red_imp_xor_eq)

  apply (erule_tac x=t in allE) back
  apply auto
  apply (erule_tac x="t'" in allE)
  apply auto
  
  apply (rule_tac x="t'a" in exI)
  apply auto
  apply (erule xor_eq.trans)
  apply auto

  apply (subgoal_tac "t \<approx> ZERO")
  apply force
  apply (rule xor_eq.symm)
  apply (rule xor_eq.trans) prefer 2
  apply assumption
  apply (erule xor_eq.symm)
  done

lemma ffactors_normed:
  "\<lbrakk> t \<in> ffactors s; normed s \<rbrakk> \<Longrightarrow> normed t"
  apply (frule  ffactors_fsubterms)
  apply force
  apply (drule normed_fsubterms)
  by auto


lemma normed_xoreq: "\<lbrakk> x \<approx> y; normed x; normed y \<rbrakk> \<Longrightarrow> x = y"
  apply (drule equiv_imp_norm)
  apply (simp add: norm_normed_id)
  done

lemma factors_Xor: "A \<in> factors (Xor X Y)
                    \<Longrightarrow> A \<in> factors X \<or> A \<in> factors Y \<or> A = Zero"
  apply (subgoal_tac "(norm (XOR (Rep_msg X) (Rep_msg Y)) \<in> msg)")
  prefer 2
  apply (simp add: msg_def normed_norm del: norm.simps)
  apply (auto simp add: Xor_def factors_def Abs_msg_inverse Zero_def simp del: norm.simps)
  apply (rule_tac x=a in exI, auto simp del: norm.simps)
  apply (subgoal_tac "(Rep_msg X \<oplus> Rep_msg Y) ~> norm (Rep_msg X \<oplus> Rep_msg Y)")
  apply (drule ffactors_xor_red)
  apply (auto simp del: norm.simps) prefer 2
  apply (rule norm_reduce)
  apply (erule_tac x="a" in allE)
  apply (erule_tac x="a" in allE)
  apply (auto simp del: norm.simps)

  apply (subgoal_tac "t' = a")
  apply force
  apply (erule normed_xoreq)
  apply (erule ffactors_normed)
  apply force
  apply (erule ffactors_normed)
  apply force

  apply (subgoal_tac "t' = a")
  apply force
  apply (erule normed_xoreq)
  apply (erule ffactors_normed)
  apply force
  apply (erule ffactors_normed)
  apply force

  apply (subgoal_tac "a = ZERO")
  apply force
  apply (erule normed_xoreq)
  apply (erule ffactors_normed)
  apply force
  apply force
  done

lemma Zero_MPair_ineq: "Zero \<noteq> MPair x y"
by (auto simp add: constructor_defs dest!: Abs_eq_normed)

declare Zero_MPair_ineq[iff]
declare  Zero_MPair_ineq[symmetric,iff]

lemma factors_Xor_Crypt:
   "Xor X Y = Crypt k m \<Longrightarrow> Crypt k m \<in> factors X \<or> Crypt k m \<in> factors Y"
  apply (subgoal_tac "Crypt k m \<in> factors (Xor X Y)") prefer 2
  apply (force)
  apply (drule factors_Xor)
  apply auto
  done

lemma factors_Xor_MPair:
   "Xor X Y = \<lbrace> A, B \<rbrace>  \<Longrightarrow> \<lbrace> A, B \<rbrace>  \<in> factors X \<or> \<lbrace> A, B \<rbrace>  \<in> factors Y"
  apply (subgoal_tac "\<lbrace> A, B \<rbrace> \<in> factors (Xor X Y)") prefer 2
  apply (force)
  apply (drule factors_Xor)
  apply (auto)
  done

lemma factors_Xor_Nonce:
   "Xor X Y = Nonce A NA  \<Longrightarrow> Nonce A NA  \<in> factors X \<or> Nonce A NA  \<in> factors Y"
  apply (subgoal_tac "Nonce A NA \<in> factors (Xor X Y)") prefer 2
  apply (force)
  apply (drule factors_Xor)
  by auto


lemma factors_Xor_Hash:
   "Xor X Y = Hash A  \<Longrightarrow> Hash A \<in> factors X \<or> Hash A  \<in> factors Y"
  apply (subgoal_tac "Hash A \<in> factors (Xor X Y)") prefer 2
  apply (force)
  apply (drule factors_Xor)
  by auto


lemma factors_LowHam:
  "\<lbrakk> d \<in> LowHamXor; x \<in> factors d \<rbrakk> \<Longrightarrow> x \<in> (range Agent \<union> {Zero} \<union> range Number \<union> range Real)"
  apply (induct rule: LowHamXor.induct)
  apply auto
  apply (drule factors_Xor)
  apply auto
  done

lemma out_context_distort:
  "\<lbrakk> d \<in> LowHamXor; out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) (Xor m d) \<rbrakk>
   \<Longrightarrow> out_context (Nonce B NB) (Hash \<lbrace>Nonce B NB, Agent B\<rbrace>) m"
  apply (drule out_context_inverse)
  apply auto
  apply (frule factors_Xor_Nonce)
  apply auto
  apply (case_tac "m = Nonce B NB")
     apply force
  apply (rule out_context.Xor) prefer 2
  apply assumption
  apply force
  apply force
  apply force
  apply (auto dest: factors_LowHam)

  apply (frule factors_Xor_Hash)
  apply auto
  apply (case_tac "m = Hash X")
    apply force
  apply (rule out_context.Xor) prefer 2
  apply assumption
  apply force
  apply force
  apply force
  apply (auto dest: factors_LowHam)

  apply (frule factors_Xor_Crypt)
  apply auto
  apply (case_tac "m = Crypt k X")
    apply force
  apply (rule out_context.Xor) prefer 2
  apply assumption
  apply force
  apply force
  apply force
  apply (auto dest: factors_LowHam)

  apply (frule factors_Xor_MPair)
  apply auto
  apply (case_tac "m = \<lbrace>X, Y\<rbrace>")
    apply force
  apply (rule out_context.Xor) prefer 2
  apply assumption
  apply force
  apply force
  apply force
  apply (auto dest: factors_LowHam)

  apply (frule factors_Xor_MPair)
  apply auto
  apply (case_tac "m = \<lbrace>X, Y\<rbrace>")
    apply force
  apply (rule out_context.Xor) prefer 2
  apply assumption
  apply force
  apply force
  apply force
  apply (auto dest: factors_LowHam)


  apply (drule factors_Xor)
  apply auto prefer 3
  apply (drule out_context_inverse)
  apply auto

  apply (case_tac "X = m", auto)
  apply (case_tac "X = m", auto)
  apply (drule factors_LowHam)
  apply assumption
  apply (drule out_context_inverse)
  apply auto
  done

lemma ffactors_not_xor:
  "x \<in> ffactors y \<Longrightarrow> {x} = ffactors x"
  apply (induct y arbitrary: x)
  apply auto
  done

lemma factors_not_xor:
  "x \<in> factors y \<Longrightarrow> factors x = {x}"
  apply (simp add: factors_def)
  apply (elim exE conjE)
  apply (subgoal_tac "a \<in> msg") prefer 2
  apply (simp add: msg_def)
  apply (rule ffactors_normed)
  apply force
  apply force
  apply clarsimp
  apply (simp add: Abs_msg_inverse)
  apply (drule ffactors_not_xor [THEN sym])
  apply auto
  done

lemma Xor_ZeroL[simp]:  "Xor Zero a = a"
  apply (auto simp add: Xor_def Zero_def)
  apply (subgoal_tac "normed ZERO")
  apply (simp add: Abs_msg_inverse)
  apply auto
  apply (subgoal_tac "normed (Rep_msg a)") prefer 2
  apply force
  apply (simp add: norm_normed_id)
  apply (simp add: Rep_msg_inverse)
done


lemma ffactors_Zero_imp_Zero:
  "\<lbrakk>  normed X; ZERO \<in> ffactors X \<rbrakk> \<Longrightarrow> X = ZERO"
  apply (induct rule: normed.induct)
  apply auto
  done

lemma factors_Zero_imp_Zero:
  "Zero \<in> factors X \<Longrightarrow> X = Zero"
  apply (auto simp add: factors_def Zero_def)
  apply (subgoal_tac "ZERO \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "a \<in> msg") prefer 2
  apply (drule ffactors_normed)
  apply force
  apply (force simp add: msg_def)
  apply (simp add: Abs_msg_inject)
  apply (subgoal_tac "normed (Rep_msg X)")
  apply (drule ffactors_Zero_imp_Zero)
  apply auto
  apply (subgoal_tac "Abs_msg (Rep_msg X) = Abs_msg ZERO") prefer 2
  apply force
  apply (subgoal_tac "Abs_msg (Rep_msg X) = X")
  apply force
  apply (rule Rep_msg_inverse)
  done

lemma n:
  "\<lbrakk> normed a;
     normed b;
     standard a \<or> standard b;
     (ffactors a \<inter> ffactors b) = {};
     ZERO \<notin> ffactors a \<union> ffactors b \<rbrakk>
   \<Longrightarrow> ffactors (normxor a b) = ffactors a \<union> ffactors b \<and> normxor a b \<noteq> ZERO"
proof (induct a  arbitrary: b rule: normed.induct)
  case (Agent aa)
  thus ?case
    apply (induct b rule: normed.induct)
    apply (case_tac "aa = a")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done
next
  case (Xor u v)
  from `normed b` show ?case using prems proof (induct b rule: normed.induct)
    case (Agent a)
    thus ?case apply -
      apply clarsimp
      done
  next
    case (Xor f g)
    from `standard (u \<oplus> v) \<or> standard (f \<oplus> g)` show ?case by auto
  next
    case Zero
    from `ZERO \<notin> ffactors (u \<oplus> v) \<union> ffactors ZERO` show ?case by auto
  next
    case (Nonce c d)
    thus ?case by clarsimp
  next
    case (Key k)
    thus ?case by clarsimp
  next
    case (Hash h)
    thus ?case by clarsimp
  next
    case (MPair f g)
    thus ?case by clarsimp
  next
    case (Crypt k m)
    thus ?case by clarsimp
  next
    case (Real m)
    thus ?case by clarsimp
  next
    case (Number n)
    thus ?case by clarsimp
  qed
next
  case (Number aa)
  thus ?case
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (case_tac "aa = n")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done 
next
  case (Hash hh)
  from `normed b` show ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "hh = h")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done 
next
  case (Key kk)
  thus ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "kk = k")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done
next
  case (Real rr)
  thus ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "rr = r")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done
next
  case (Nonce nn aa)
  thus ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "NONCE nn aa = NONCE n t")
    apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done
next
  case Zero
  from `ZERO \<notin> ffactors ZERO \<union> ffactors b` show ?case by auto
next
  case (Crypt kk mm)
  from `normed b` show ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "CRYPT mm kk = CRYPT k m")
      apply simp
    apply (force split: split_if_asm)
    apply clarsimp
    done
next
  case (MPair aa bb)
  from `normed b` show ?case using prems
    apply (induct b rule: normed.induct)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply (case_tac "MPAIR aa bb = MPAIR a b")
      apply simp
    apply (force split: split_if_asm)
    apply (force split: split_if_asm)
    apply clarsimp
    done
qed

lemma m:
  "\<lbrakk> normed X;
     NONCE A NA \<notin> ffactors X;
     ZERO \<notin> ffactors X
   \<rbrakk> \<Longrightarrow> ffactors (X \<otimes> NONCE A NA) = (ffactors X \<union> ffactors (NONCE A NA)) \<and> X \<otimes> (NONCE A NA) \<noteq> ZERO"
apply (rule n)
apply force
apply force
apply force
apply auto
done

lemma ffactors_Xor_nonce_not_subterm:
  "\<lbrakk> normed X; NONCE P NP \<notin> ffactors X \<rbrakk> \<Longrightarrow>
      (ffactors (ZERO \<otimes> (NONCE P NP)) = {NONCE P NP} \<and> X = ZERO)
    \<or> ffactors (X \<otimes> (NONCE P NP)) = {NONCE P NP} \<union> ffactors X"
apply (case_tac "X=ZERO")
apply clarsimp
apply (case_tac "ZERO \<in> ffactors X")
apply (drule  ffactors_Zero_imp_Zero)
apply force
apply force
apply (frule m)
apply simp
apply simp
apply simp
done

lemma factors_Xor_nonce_not_subterm:
  "\<lbrakk> Nonce P NP \<notin> factors X \<rbrakk> \<Longrightarrow>
      (factors (Xor Zero (Nonce P NP)) = {Nonce P NP} \<and> X = Zero)
    \<or> factors (Xor X (Nonce P NP)) = {Nonce P NP} \<union> factors X"
  apply (unfold factors_def Xor_def Zero_def Nonce_def)
  apply (subgoal_tac "ZERO \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (subgoal_tac "Rep_msg  X \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (subgoal_tac "NONCE P NP \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (simp add: Abs_msg_inverse)
  apply (subgoal_tac "normed (Rep_msg X)") prefer 2
    apply force
    apply (drule_tac P=P and NP=NP in ffactors_Xor_nonce_not_subterm)
  apply force
  apply clarsimp
  apply (elim disjE)
  apply clarsimp
  apply (drule_tac f=Abs_msg in HOL.arg_cong)
  apply (simp add: Rep_msg_inverse)
  apply (subgoal_tac "(Rep_msg X \<otimes> NONCE P NP) \<in> msg") prefer 2
  apply (simp add: msg_def)
  apply (rule normed_normxor)
  apply simp
  apply simp
   apply (simp add: Abs_msg_inverse)
   apply force
   done

lemma hash_ffactors:
  "\<lbrakk> normed X;
     normed (HASH Y);
     HASH Y \<notin> ffactors X;
     ZERO \<notin> ffactors X
   \<rbrakk> \<Longrightarrow> ffactors (X \<otimes> HASH Y) = (ffactors X \<union> ffactors (HASH Y)) \<and> X \<otimes> (HASH Y) \<noteq> ZERO"
apply (rule n)
apply force
apply force
apply force
apply auto
done

lemma ffactors_Xor_hash_not_subterm:
  "\<lbrakk> normed X; normed (HASH Y); HASH Y \<notin> ffactors X \<rbrakk> \<Longrightarrow>
      (ffactors (ZERO \<otimes> (HASH Y)) = {HASH Y} \<and> X = ZERO)
    \<or> ffactors (X \<otimes> (HASH Y)) = {HASH Y} \<union> ffactors X"
apply (case_tac "X=ZERO")
apply clarsimp
apply (case_tac "ZERO \<in> ffactors X")
apply (drule  ffactors_Zero_imp_Zero)
apply force
apply force
apply (frule hash_ffactors)
apply simp
apply simp
apply simp
apply simp
done

lemma factors_Xor_hash_not_subterm:
  "\<lbrakk> Hash Y \<notin> factors X \<rbrakk> \<Longrightarrow>
      (factors (Xor Zero (Hash Y)) = {Hash Y} \<and> X = Zero)
    \<or> factors (Xor X (Hash Y)) = {Hash Y} \<union> factors X"
  apply (unfold factors_def Xor_def Zero_def Hash_def)
  apply (subgoal_tac "ZERO \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (subgoal_tac "Rep_msg  X \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (subgoal_tac "HASH (Rep_msg Y) \<in> msg") prefer 2
    apply (force simp add: msg_def)
  apply (simp add: Abs_msg_inverse)
  apply (subgoal_tac "normed (Rep_msg X)") prefer 2
    apply force
    apply (drule_tac Y="Rep_msg Y" in ffactors_Xor_hash_not_subterm)
  apply force
  apply clarsimp
  apply (elim disjE)
  apply clarsimp
  apply (drule_tac f=Abs_msg in HOL.arg_cong)
  apply (simp add: Rep_msg_inverse)
  apply (subgoal_tac "(Rep_msg X \<otimes> HASH (Rep_msg Y)) \<in> msg") prefer 2
  apply (simp add: msg_def)
  apply (rule normed_normxor)
  apply simp
  apply simp
  apply (simp add: Abs_msg_inverse)
  apply force
  done

lemma out_context_not[dest]:
  "(out_context (Nonce (Honest P) NP) (Hash \<lbrace>Nonce (Honest P) NP, Agent (Honest P)\<rbrace>)
      (Hash \<lbrace>Nonce (Honest P) NP, Agent (Honest P)\<rbrace>)) \<Longrightarrow> False"
  apply (drule out_context_inverse)
  apply auto
  done

lemma subterms_Nonce_Nonce:
  "Nonce (Honest A) NA \<noteq> Nonce (Honest B) NB
   \<Longrightarrow> Nonce (Honest A) NA \<in> subterms {Xor (Nonce (Honest A) NA) (Nonce (Honest B) NB)}"
  apply (rule factors_imp_subterms)
  apply (simp add: Xor_comm[where A="Nonce (Honest A) NA"])
  apply (subgoal_tac "Nonce (Honest A) NA \<notin> factors (Nonce (Honest B) NB)")
  apply (drule factors_Xor_nonce_not_subterm)
  apply auto
  done

lemma subterms_xor_nonce_hash:
  "subterms {Xor (Nonce B NB) (Hash m)}
   = insert (Xor (Nonce B NB) (Hash m))
       (insert (Nonce B NB) (subterms {Hash m}))"
  apply (simp only: Nonce_def Hash_def Xor_def subterms_def)
  apply (subgoal_tac "NONCE B NB \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "HASH (Rep_msg m) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "norm
                   (Rep_msg (Abs_msg (NONCE B NB)) \<oplus>
                    Rep_msg (Abs_msg (HASH (Rep_msg m)))) \<in> msg") prefer 2
  apply (simp only: msg_def)
  apply (simp del: norm.simps)
  apply (auto simp add: Abs_msg_inverse)
  apply (rule_tac x=ma in exI)
  apply (auto split: split_if_asm)
  apply (drule fsubterms.singleton)
  apply auto
  apply (rule_tac x=ma in exI)
  apply (auto split: split_if_asm)
  apply (drule_tac H="{Rep_msg m, NONCE B NB}" in fsubterms.trans)
  apply force
  apply force
done


lemma components_MPair[simp]:
  "components {MPair a b} = components {a} \<union> components {b}"
  apply (subgoal_tac "MPAIR (Rep_msg a) (Rep_msg b) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "(Rep_msg a) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "(Rep_msg b) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: MPair_def components_def)
  apply (auto simp add: Abs_msg_inverse)
  done

lemma components_non_pair:
  "\<forall> X Y. m \<noteq> MPair X Y \<Longrightarrow> components {m} = {m}"
  apply (subgoal_tac "Rep_msg m \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (simp add: components_def MPair_def)
  apply (case_tac "Rep_msg m")
  apply auto
  apply (auto dest: HOL.arg_cong[where f=Abs_msg] simp add: Rep_msg_inverse)
  apply (drule HOL.arg_cong[where f=Abs_msg])
  apply (auto simp add: Rep_msg_inverse)
  apply (erule_tac x="Abs_msg fmsg1" in allE)
  apply (erule_tac x="Abs_msg fmsg2" in allE)
  apply (subgoal_tac "fmsg1 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (subgoal_tac "fmsg2 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (auto simp add: Abs_msg_inverse)
  
  apply (drule HOL.arg_cong[where f=Abs_msg])
  apply (auto simp add: Rep_msg_inverse)
  apply (erule_tac x="Abs_msg fmsg1" in allE)
  apply (erule_tac x="Abs_msg fmsg2" in allE)
  apply (subgoal_tac "fmsg1 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (subgoal_tac "fmsg2 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (auto simp add: Abs_msg_inverse)

  apply (drule HOL.arg_cong[where f=Abs_msg])
  apply (auto simp add: Rep_msg_inverse)
  apply (erule_tac x="Abs_msg fmsg1" in allE)
  apply (erule_tac x="Abs_msg fmsg2" in allE)
  apply (subgoal_tac "fmsg1 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (subgoal_tac "fmsg2 \<in> msg") prefer 2
  apply (force simp add: msg_def elim: normed_MPAIR)
  apply (auto simp add: Abs_msg_inverse)
  done


lemma components_nonce[simp]:
  "components {Nonce A NA} = {Nonce A NA}"
  by (rule components_non_pair, auto)

lemma components_crypt[simp]:
  "components {Crypt k m} = {Crypt k m}"
  by (rule components_non_pair, auto)

lemma components_hash[simp]:
  "components {Hash m} = {Hash m}"
  by (rule components_non_pair, auto)

lemma components_xor_n_n_a:
  "components {Xor (Nonce A NA) (Xor (Nonce B NB) (Agent C))}
   = {Xor (Nonce A NA) (Xor (Nonce B NB) (Agent C))}"
  apply (rule components_non_pair)
  apply (subgoal_tac "NONCE A NA \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "NONCE B NB \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "AGENT C \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp add: Xor_def MPair_def Nonce_def Agent_def simp del: norm.simps)
  apply (subgoal_tac "MPAIR (Rep_msg X) (Rep_msg Y) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "normed (norm
         (Rep_msg (Abs_msg (NONCE A NA)) \<oplus>
           norm
          (Rep_msg (Abs_msg (NONCE B NB)) \<oplus> Rep_msg (Abs_msg (AGENT C)))))") prefer 2
  apply (force simp add: msg_def)
  
  apply (auto simp add: Abs_msg_inverse  split: split_if_asm)
  apply (auto simp add: Abs_msg_inject XORnz_def)
  done

(* move *)
lemma Key_parts_Xor[dest]:
  "Key k \<in> parts {Xor X Z} \<Longrightarrow> Key k \<in> parts {X, Z}"
  apply (auto simp add: Key_def parts_def)
  apply (drule_tac f="Rep_msg" in HOL.arg_cong)
  apply (subgoal_tac "normed m") prefer 2
  apply (erule normed_fparts)
  apply auto
  apply (subgoal_tac "normed (KEY k)")
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "normed (norm (Rep_msg X \<oplus> Rep_msg Z))") prefer 2
  apply force
  apply (simp only: Abs_msg_normed)
  apply (drule fparts_norm_KEY)
by auto

lemma Xor_same_arg:
  assumes P: "Xor a b = Xor a c"
  shows  "b = c"
proof -
  have A: "b = Xor (Xor a b) a" by (simp add: Xor_rewrite)
  have B: "Xor (Xor a c) a = c" by (simp add: Xor_rewrite)
  show ?thesis using P apply -
    apply (subst A)
    apply (subst P)
    apply (subst B)
    by simp
qed

(* MOVE *)
lemma sig_subterms:
  "Crypt k M \<in> subterms {Xor X Y}
  \<Longrightarrow> Crypt k M \<in> subterms {X, Y}"
  apply (auto simp add: Crypt_def subterms_def MPair_def)
  apply (subgoal_tac "Rep_msg M \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (subgoal_tac "CRYPT k (Rep_msg M) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (drule_tac f="Rep_msg" in HOL.arg_cong)
  apply (subgoal_tac "normed m") prefer 2
  apply (erule normed_fsubterms)
  apply force
  apply auto
  apply (simp add: Abs_msg_inverse)
  apply (rule_tac x="CRYPT k (Rep_msg M)" in exI)
  apply auto
  apply (unfold Xor_def)
  apply (subgoal_tac "normed (norm (Rep_msg X \<oplus> Rep_msg Y))") prefer 2
  apply (rule normed_norm)
  apply (simp only: Abs_msg_normed)
  apply (drule fsubterms_norm_CRYPT)
  apply auto
  apply (subgoal_tac "m' = Rep_msg M")
  apply force
  apply (subgoal_tac "normed m'")
  apply (simp only: norm_normed_id)
  apply (subgoal_tac "m' \<in> fsubterms {Rep_msg X, Rep_msg Y}") prefer 2
  apply (rule_tac G="{CRYPT k m'}" in fsubterms.trans)
  apply force
  apply force
  apply (drule_tac X="m'" in fsubterms.singleton)
  apply auto
  apply (drule_tac Y="m'" in normed_fsubterms, auto)
  apply (drule_tac Y="m'" in normed_fsubterms, auto)
done

lemma parts_in_subterms:
  "x \<in> parts S \<Longrightarrow> x \<in> subterms S"
  apply (unfold parts_def subterms_def)
  apply (auto)
  apply (rule_tac x=m in exI)
  apply auto
  apply (erule fparts.induct)
  apply (auto intro: fsubterms.Inj fsubterms.Fst fsubterms.Snd fsubterms.Ctext fsubterms.Hash
              fsubterms.Xor1 fsubterms.Xor2
              dest: fparts_fsubterms_Abs_msg)
  done

  
lemma subterms_component_trans:
  "\<lbrakk> X \<in> subterms{Y}; Y \<in> components {Z} \<rbrakk> \<Longrightarrow> X \<in> subterms {Z}"
  apply (rule_tac subterms.trans)
  apply simp
  apply (drule components_subset_parts)
  apply auto
  apply (erule parts_in_subterms)
  done

lemma xor_nz[simp]: "b \<noteq> ZERO \<Longrightarrow> a \<odot> b = a \<oplus> b"
  apply (case_tac b)
  apply auto
done

lemma fsubterms_xor_nonce_right:
  "\<lbrakk> normed b;
     normed a;
     NONCE A NA \<in> fsubterms {b}; 
     NONCE A NA \<notin> fsubterms {a} \<rbrakk>
   \<Longrightarrow> NONCE A NA \<in> fsubterms {norm (a \<oplus> HASH b)}"
apply (auto simp add: norm_normed_id)
apply (induct a)
apply (auto split: split_if_asm)
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force) defer
apply (erule fsubterms.trans, force)
apply (rule_tac G="{HASH b}" in fsubterms.trans)
apply force
apply force
apply (erule fsubterms.trans, force)
apply (rule_tac G="{HASH b}" in fsubterms.trans)
apply force
apply force
apply (erule fsubterms.trans, force)
apply (erule fsubterms.trans, force)
apply (drule_tac H="{b,a2}" in fsubterms.trans)
apply force
apply force defer
apply (rule_tac G="{HASH b}" in fsubterms.trans)
apply force
apply force
apply (subgoal_tac "normed a2") prefer 2
apply (rule normed_xor_snd)
apply force
apply (subgoal_tac "NONCE A NA \<notin> fsubterms {a2}") prefer 2
apply clarsimp
apply (drule_tac H="{a1,a2}" in fsubterms.trans) back
apply force
apply force
apply auto
apply (subst xor_nz)
apply force
apply auto
apply (erule_tac fsubterms.trans)
apply force
done


lemma subterms_xor_nonce_right:
  "\<lbrakk> Nonce A NA \<notin> subterms {a} \<rbrakk>
   \<Longrightarrow> Nonce A NA \<in> subterms {Xor a (Hash \<lbrace> Nonce A NA, Agent B \<rbrace>)}"
  apply (auto simp del: norm.simps simp add: subterms_def Xor_def Hash_def Nonce_def Agent_def MPair_def)
  apply (rule_tac x="NONCE A NA" in exI)
  apply (auto simp del: norm.simps)
  apply (subgoal_tac "AGENT B \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp only: Abs_msg_inverse)
  apply (subgoal_tac "NONCE A NA \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp only: Abs_msg_inverse)
  apply (subgoal_tac "MPAIR (NONCE A NA) (AGENT B) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp only: Abs_msg_inverse)
  apply (subgoal_tac "HASH (MPAIR (NONCE A NA) (AGENT B)) \<in> msg") prefer 2
  apply (force simp add: msg_def)
  apply (auto simp only: Abs_msg_inverse)
  apply (rule fsubterms_xor_nonce_right)
  apply auto
done


end