(******************************************************************************

  ProtoVeriPhy

  Module: distance/NatTree.thy

  Copyright (C) 2008-2012 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Tree with Nat labeled nodes and *}

theory NatTree imports Main begin

datatype
  'leaf tree =  Leaf nat "'leaf"
             |  Node nat "('leaf tree)" "('leaf tree)"

subsection {* Linear Order on trees *}

instantiation tree :: (linorder) linorder
begin

fun
  less_tree :: "'a tree \<Rightarrow> 'a tree \<Rightarrow> bool"
 where
  "(Leaf a x) <  (Leaf b y)  = (if (a = b) then x < y else a < b)" |
  "(Node a n1 n2) < (Node b m1 m2) = (if (a = b)
                                        then (if (n1 = m1) then n2 < m2 else n1 < m1)
                                        else (a < b))" |
   "(Leaf _ _) < (Node _ _ _) = True" |
   "(Node _ _ _) < (Leaf _ _) = False"

definition less_eq_tree: "(a::'a tree) \<le> b = ((a = b) \<or> (a < b))"

lemma antisym2: "(x :: 'a tree) < y \<Longrightarrow> \<not> y < x"
  apply (induct arbitrary: y rule: tree.induct)
  apply (case_tac y, auto)
  apply (case_tac y, auto split: split_if_asm)
done

lemma antisym:
  fixes x y :: "'a tree" shows "(x < y) = (x \<le> y \<and> \<not> y \<le> x)"
proof -
  have "\<not> x < x" by (induct rule: tree.induct, auto)
  thus ?thesis using antisym2 by (auto simp add: less_eq_tree)
qed

instance proof
  fix x y :: "'a tree" show "(x < y) = (x \<le> y \<and> \<not> y \<le> x)" using antisym by auto
next
  fix x :: "'a tree" show "x \<le> x" by (auto simp add: less_eq_tree)
next
  fix x y :: "'a tree" show "\<lbrakk>x \<le> y; y \<le> x\<rbrakk> \<Longrightarrow> x = y"
    apply (insert antisym [of x y])
    apply (unfold less_eq_tree)
    by clarsimp
next
  fix x y :: "'a tree" show "x \<le> y \<or> y \<le> x"
    apply (induct arbitrary: y rule: tree.induct)
    apply (auto simp add: less_eq_tree)
    apply (case_tac y, auto split: split_if_asm)
    apply (case_tac y, auto split: split_if_asm)
    by force
next
  fix x y z :: "'a tree" show "\<lbrakk>x \<le> y; y \<le> z\<rbrakk> \<Longrightarrow> x \<le> z"
  proof (induct arbitrary: x z rule: tree.induct)
    case (Leaf nat leaf x z)
      thus ?case
	apply (case_tac z, case_tac x) prefer 3
	apply (case_tac x)
	apply (auto simp add: less_eq_tree split: split_if_asm)
	done
  next
    case (Node nat tree1 tree2 x z) thus ?case
    proof (cases z)
      case (Leaf n t) thus ?thesis using prems(3-)
	by (auto simp add: less_eq_tree split: split_if_asm)
    next
      case (Node natz tree1z tree2z) thus ?thesis
      proof (cases x)
	case (Leaf n t) thus ?thesis using prems(3-)
	  by (auto simp add: less_eq_tree split: split_if_asm)
      next
	case (Node natx tree1x tree2x) thus ?thesis
	proof (cases)
	  assume "natx = natz" thus ?thesis
	    proof -
	      have t1: "\<lbrakk>tree1x \<le> tree1; tree1 \<le> tree1z\<rbrakk> \<Longrightarrow> tree1x \<le> tree1z" using prems(1) .
	      have t2: "\<lbrakk>tree2x \<le> tree2; tree2 \<le> tree2z\<rbrakk> \<Longrightarrow> tree2x \<le> tree2z" using prems(2) .
	      show ?thesis using t1 t2 prems(3-)
		apply (auto simp add: less_eq_tree split: split_if_asm)
		by (auto dest: antisym2)
	    qed
	  next
	  assume "natx \<noteq> natz" thus ?thesis using prems(3-)
	    by (auto simp add: less_eq_tree split: split_if_asm)
	qed
      qed
    qed
  qed
qed

end

end